from dataclasses import dataclass
from typing import Mapping
from hugedict.misc import identity
from itertools import chain
from kgdata.wikidata import db
from kgdata.wikidata.models import WDEntity, WDValue, WDClass, WDProperty
from sm.namespaces.wikidata import WikidataNamespace
from sand.models.base import StoreWrapper
from sand.models.entity import Entity, Statement, Value
from sand.models.ontology import (
    DEFAULT_ONT_PROPS,
    OntClass,
    OntProperty,
    DEFAULT_ONT_CLASSES,
    OntPropertyDataType,
)


wdns = WikidataNamespace.create()
WD_ONT_CLASSES = {
    wdns.get_rel_uri(wdns.STATEMENT_URI): OntClass(
        id=wdns.get_rel_uri(wdns.STATEMENT_URI),
        uri=wdns.STATEMENT_URI,
        label=wdns.get_rel_uri(wdns.STATEMENT_URI),
        aliases=[],
        description="",
        parents=[],
    )
}
WD_ONT_CLASSES.update(DEFAULT_ONT_CLASSES)
INVERSE_DEFAULT_URI2ID = {
    v.uri: k for k, v in chain(DEFAULT_ONT_PROPS.items(), WD_ONT_CLASSES.items())
}

WD_DATATYPE_MAPPING: Mapping[str, OntPropertyDataType] = {
    "monolingualtext": "monolingualtext",
    "url": "url",
    "wikibase-item": "entity",
    "external-id": "entity",
    "time": "datetime",
    "quantity": "number",
    "string": "string",
    "globe-coordinate": "globe-coordinate",
}


@dataclass
class WrapperWDEntity(Entity):
    @property
    def readable_label(self):
        return f"{self.label} ({self.id})"

    @property
    def instanceof(self):
        """Get the property representing the instanceof relation."""
        return "P31"


@dataclass
class WrapperWDClass(OntClass):
    @property
    def readable_label(self):
        return f"{self.label} ({self.id})"


@dataclass
class WrapperWDProperty(OntProperty):
    @property
    def readable_label(self):
        return f"{self.label} ({self.id})"


def get_entity_db(dbfile: str, proxy: bool):
    store = db.get_entity_db(dbfile, proxy=proxy, read_only=not proxy)
    return StoreWrapper(
        store,
        key_deser=identity,
        val_deser=qnode_deser,
    )


def get_ontclass_db(dbfile: str, proxy: bool):
    return StoreWrapper(
        db.get_wdclass_db(dbfile, proxy=proxy, read_only=not proxy),
        key_deser=get_wdclass_id,
        val_deser=ont_class_deser,
    )


def get_ontprop_db(dbfile: str, proxy: bool):
    return StoreWrapper(
        db.get_wdprop_db(dbfile, proxy=proxy, read_only=not proxy),
        key_deser=get_wdprop_id,
        val_deser=ont_prop_deser,
    )


def qnode_deser(qnode: WDEntity):
    props = {}
    for pid, stmts in qnode.props.items():
        new_stmts = []
        for stmt in stmts:
            new_stmt = Statement(
                value=wd_value_deser(stmt.value),
                qualifiers={
                    qid: [wd_value_deser(x) for x in lst]
                    for qid, lst in stmt.qualifiers.items()
                },
                qualifiers_order=stmt.qualifiers_order,
            )
            new_stmts.append(new_stmt)
        props[pid] = new_stmts

    return WrapperWDEntity(
        id=qnode.id,
        label=qnode.label,
        aliases=qnode.aliases,
        description=qnode.description,
        properties=props,
    )


def ont_class_deser(item: WDClass):
    return WrapperWDClass(
        id=item.id,
        uri=WikidataNamespace.get_entity_abs_uri(item.id),
        aliases=item.aliases,
        label=item.label,
        description=item.description,
        parents=item.parents,
        ancestors=item.ancestors,
    )


def ont_prop_deser(item: WDProperty):
    global WD_DATATYPE_MAPPING
    return WrapperWDProperty(
        id=item.id,
        uri=WikidataNamespace.get_prop_abs_uri(item.id),
        aliases=item.aliases,
        label=item.label,
        description=item.description,
        datatype=WD_DATATYPE_MAPPING[item.datatype] if item.datatype in WD_DATATYPE_MAPPING else "unknown",
        parents=item.parents,
        ancestors=item.ancestors,
    )


def wd_value_deser(val: WDValue):
    if val.is_entity_id(val):
        return Value(type="entityid", value=val.as_entity_id())
    elif val.is_string(val):
        return Value(type="string", value=val.as_string())
    else:
        return Value(type=val.type, value=val.value)


def get_wdprop_id(uri_or_id: str):
    if uri_or_id.startswith("http"):
        return uri_or_id.replace(f"http://www.wikidata.org/prop/", "")
    return uri_or_id


def get_wdclass_id(uri_or_id: str):
    if uri_or_id.startswith("http"):
        return uri_or_id.replace(f"http://www.wikidata.org/entity/", "")
    return uri_or_id


def uri2id(uri: str):
    if uri.startswith("http://www.wikidata.org/prop/"):
        return WikidataNamespace.get_prop_id(uri)
    if uri.startswith("http://www.wikidata.org/entity/"):
        return WikidataNamespace.get_entity_id(uri)
    if uri in INVERSE_DEFAULT_URI2ID:
        return INVERSE_DEFAULT_URI2ID[uri]
    return uri


def get_rel_uri(uri: str):
    return wdns.get_rel_uri(uri)



import os

import sm.outputs as O
from gena import generate_api, generate_app, generate_readonly_api_4dict
from hugedict.chained_mapping import ChainedMapping
from sm.misc.funcs import import_attr

from sand.config import SETTINGS
from sand.controllers.assistant import assistant_bp
from sand.controllers.project import project_bp
from sand.controllers.table import table_bp, table_row_bp
from sand.controllers.settings import setting_bp
from sand.deserializer import deserialize_graph
from sand.models import EntityAR, SemanticModel
from sand.models.ontology import OntClass, OntClassAR, OntProperty, OntPropertyAR
from sand.serializer import (
    batch_serialize_sms,
    serialize_class,
    serialize_entity,
    serialize_property,
)

app = generate_app(
    [
        table_bp,
        project_bp,
        assistant_bp,
        table_row_bp,
        setting_bp,
        generate_api(
            SemanticModel,
            deserializers={"data": deserialize_graph},
            batch_serialize=batch_serialize_sms,
        ),
        generate_readonly_api_4dict(
            "entities",
            serialize=serialize_entity,
            id2ent=ChainedMapping(
                EntityAR(), import_attr(SETTINGS["entity"]["default"])
            ),
        ),
        generate_readonly_api_4dict(
            "classes",
            serialize=serialize_class,
            id2ent=ChainedMapping(
                OntClassAR(), import_attr(SETTINGS["ont_classes"]["default"])
            ),
            unique_field_funcs={"uri": OntClass.uri2id},
        ),
        generate_readonly_api_4dict(
            "properties",
            serialize=serialize_property,
            id2ent=ChainedMapping(
                OntPropertyAR(),
                import_attr(SETTINGS["ont_props"]["default"]),
            ),
            unique_field_funcs={"uri": OntProperty.uri2id},
        ),
    ],
    os.path.dirname(__file__),
)

app.config["MAX_CONTENT_LENGTH"] = 16 * 1024 * 1024  # maximum upload of 16 MB



from __future__ import annotations
from dataclasses import asdict, dataclass
from typing import Optional, NamedTuple, List, Dict, Tuple, Union
from gena.custom_fields import (
    ListDataClassField,
    DataClassField,
    Dict2ListDataClassField,
)
from grams.inputs.context import ContentHierarchy

import orjson
from peewee import CharField, ForeignKeyField, CompositeKey, TextField, IntegerField
from playhouse.shortcuts import model_to_dict
from playhouse.sqlite_ext import JSONField

from sand.models.base import BaseModel
from sand.models.entity import Value
from sand.models.project import Project


@dataclass
class CandidateEntity:
    entity_id: str  # entity id, can't be NIL_ENTITY
    probability: float


@dataclass
class Link:
    start: int
    end: int
    # none when the entity is not mapped yet, if there is no entity, use NIL_ENTITY
    url: Optional[str]
    entity_id: Optional[str]  # entity id
    candidate_entities: List[CandidateEntity]

    @staticmethod
    def from_tuple(value):
        return Link(*value[:-1], [CandidateEntity(*x) for x in value[-1]])  # type: ignore


@dataclass
class ContextPage:
    url: str
    title: str
    entity: Optional[str]  # entity id associated with the page


class Table(BaseModel):
    name = CharField()
    description = TextField()
    # list of columns in the table
    columns: List[str] = JSONField()  # type: ignore
    project = ForeignKeyField(Project, backref="tables", on_delete="CASCADE")
    size = IntegerField()
    context_page: Optional[ContextPage] = DataClassField(ContextPage, null=True)  # type: ignore
    context_values: List[Value] = ListDataClassField(Value)  # type: ignore
    context_tree: List[ContentHierarchy] = ListDataClassField(ContentHierarchy)  # type: ignore

    class Meta:
        indexes = ((("project", "name"), True),)

    def to_dict(self):
        return {
            "id": self.id,  # type: ignore
            "name": self.name,
            "description": self.description,
            "columns": self.columns,
            "project": self.project_id,  # type: ignore
            "size": self.size,
            "context_page": asdict(self.context_page)
            if self.context_page is not None
            else None,
            "context_values": [asdict(x) for x in self.context_values]
            if self.context_values is not None
            else None,
            "context_tree": [x.to_dict() for x in self.context_tree]
            if self.context_tree is not None
            else None,
        }


class TableRow(BaseModel):
    # fmt: off
    table = ForeignKeyField(Table, backref="rows", on_delete="CASCADE")
    index = IntegerField()  # type: ignore
    row: List[Union[str, float]] = JSONField(json_dumps=orjson.dumps, json_loads=orjson.loads)  # type: ignore
    links: Dict[str, List[Link]] = Dict2ListDataClassField(Link)  # type: ignore
    # fmt: on

    class Meta:
        indexes = ((("table", "index"), True),)

    def to_dict(self):
        dlinks = {}
        if self.links is not None:
            for ci, links in self.links.items():
                newlinks = []
                for link in links:
                    link = asdict(link)
                    # link["candidate_entities"] = [
                    #     cent._asdict() for cent in link["candidate_entities"]
                    # ]
                    newlinks.append(link)
                dlinks[ci] = newlinks

        return {
            "id": self.id,  # type: ignore
            "table": self.table_id,  # type: ignore
            "index": self.index,
            "row": self.row,
            "links": dlinks,
        }



from dataclasses import asdict
from functools import partial
from typing import Callable, Dict, List, Mapping, Optional

from playhouse.shortcuts import model_to_dict

import sm.outputs as O
from sand.models import SemanticModel, Table
from sand.models.entity import Entity, EntityAR
from sand.models.ontology import OntProperty, OntClass, OntPropertyAR, OntClassAR


def serialize_property(prop: OntProperty):
    return {
        "id": prop.id,
        "uri": prop.uri,
        "label": prop.label,
        "readable_label": prop.readable_label,
        "aliases": prop.aliases,
        "description": prop.description,
        "datatype": prop.datatype,
        "parents": prop.parents,
        "ancestors": list(prop.ancestors),
    }


def serialize_class(cls: OntClass):
    return {
        "id": cls.id,
        "uri": cls.uri,
        "label": cls.label,
        "readable_label": cls.readable_label,
        "aliases": cls.aliases,
        "description": cls.description,
        "parents": cls.parents,
        "ancestors": list(cls.ancestors),
    }


def serialize_entity(ent: Entity):
    return {
        "id": ent.id,
        "uri": Entity.id2uri(ent.id),
        "readable_label": ent.readable_label,
        "label": ent.label.to_dict(),
        "aliases": ent.aliases.to_dict(),
        "description": ent.description.to_dict(),
        "properties": {
            prop: [
                {
                    "value": asdict(stmt.value),
                    "qualifiers": {
                        qid: [asdict(qval) for qval in qvals]
                        for qid, qvals in stmt.qualifiers.items()
                    },
                    "qualifiers_order": stmt.qualifiers_order,
                }
                for stmt in stmts
            ]
            for prop, stmts in ent.properties.items()
        },
    }


def serialize_graph(
    sm: O.SemanticModel,
    uri2label: Callable[[str, bool], Optional[str]],
    columns: Optional[List[str]] = None,
):
    nodes = []
    for n in sm.iter_nodes():
        if isinstance(n, O.ClassNode):
            nodes.append(
                {
                    "id": str(n.id),
                    "uri": n.abs_uri,
                    "label": uri2label(n.abs_uri, True) or n.rel_uri,
                    "approximation": n.approximation,
                    "type": "class_node",
                }
            )
        elif isinstance(n, O.DataNode):
            nodes.append(
                {
                    "id": str(n.id),
                    "label": columns[n.col_index] if columns is not None else n.label,
                    "type": "data_node",
                    "column_index": n.col_index,
                }
            )
        else:
            if n.datatype == O.LiteralNodeDataType.Entity:
                value = {
                    "id": Entity.uri2id(n.value),
                    "uri": n.value,
                    "type": n.datatype.value,
                }
            else:
                value = {"type": n.datatype.value, "value": n.value}
            nodes.append(
                {
                    "id": str(n.id),
                    "value": value,
                    "label": n.label,
                    "type": "literal_node",
                    "is_in_context": n.is_in_context,
                }
            )

    edges = [
        {
            "source": str(e.source),
            "target": str(e.target),
            "uri": e.abs_uri,
            "label": e.readable_label or uri2label(e.abs_uri, False) or e.rel_uri,
            "approximation": e.approximation,
        }
        for e in sm.iter_edges()
    ]

    return dict(nodes=nodes, edges=edges)


def batch_serialize_sms(sms: List[SemanticModel]):
    # tbls = {sm.table_id for sm in sms if sm.table_id is not None}  # type: ignore
    # tbls = {
    #     tbl.id: tbl
    #     for tbl in Table.select(Table.id, Table.columns).where(Table.id.in_(tbls))  # type: ignore
    # }

    ontprops = OntPropertyAR()
    ontclasses = OntClassAR()
    uri2lbl = partial(get_label, ontprops=ontprops, ontclasses=ontclasses)

    output = []
    for sm in sms:
        r = model_to_dict(sm, recurse=False)
        if r["data"] is not None:
            # columns = None if sm.table_id is None else tbls[sm.table_id].columns
            columns = None
            r["data"] = serialize_graph(r["data"], uri2lbl, columns)  # type: ignore
        output.append(r)
    return output


def get_label(
    id: str,
    is_class: bool,
    ontprops: Mapping[str, OntProperty],
    ontclasses: Mapping[str, OntClass],
) -> Optional[str]:
    if is_class:
        if id in ontclasses:
            return ontclasses[id].readable_label
        elif id in ontprops:
            return ontprops[id].readable_label
    else:
        if id in ontprops:
            return ontprops[id].readable_label
        elif id in ontclasses:
            return ontclasses[id].readable_label
    return None



from dataclasses import dataclass, field
from typing import Dict, List, Literal, Mapping, Set

from rdflib import RDFS
from sm.misc.funcs import import_func
from sand.config import SETTINGS


@dataclass
class OntClass:
    id: str
    uri: str
    label: str
    aliases: List[str]
    description: str
    parents: List[str]
    ancestors: Set[str] = field(default_factory=set)

    @property
    def readable_label(self):
        return self.label

    @staticmethod
    def uri2id(uri: str) -> str:
        """Convert class URI to entity ID."""
        raise NotImplementedError(
            "The method is set when its store is initialized. Check the call order to ensure `OntClassAR` is called first"
        )


OntPropertyDataType = Literal[
    "monolingualtext",
    "url",
    "entity",
    "datetime",
    "number",
    "string",
    "globe-coordinate",
    "unknown"  # don't know the type yet
]


@dataclass
class OntProperty:
    id: str
    uri: str
    label: str
    datatype: OntPropertyDataType
    aliases: List[str]
    description: str
    parents: List[str]
    ancestors: Set[str] = field(default_factory=set)

    @property
    def readable_label(self):
        return self.label

    @staticmethod
    def uri2id(uri: str) -> str:
        """Convert property URI to entity ID."""
        raise NotImplementedError(
            "The method is set when its store is initialized. Check the call order to ensure `OntPropertyAR` is called first"
        )


PROP_AR = None
CLASS_AR = None

DEFAULT_ONT_PROPS = {
    "rdfs:label": OntProperty(
        id="rdfs:label",
        uri=str(RDFS.label),
        label="rdfs:label",
        aliases=[],
        datatype="string",
        description="",
        parents=[],
    )
}
DEFAULT_ONT_CLASSES = {}


def OntPropertyAR() -> Mapping[str, OntProperty]:
    """Get a mapping from id (not url) to the ontology property"""
    global PROP_AR

    if PROP_AR is None:
        cfg = SETTINGS["ont_props"]
        func = import_func(cfg["constructor"])
        PROP_AR = func(**cfg["args"])
        OntProperty.uri2id = import_func(cfg["uri2id"])

    return PROP_AR


def OntClassAR() -> Mapping[str, OntClass]:
    """Get a mapping from id (not url) to the ontology class"""
    global CLASS_AR

    if CLASS_AR is None:
        cfg = SETTINGS["ont_classes"]
        func = import_func(cfg["constructor"])
        CLASS_AR = func(**cfg["args"])
        OntClass.uri2id = import_func(cfg["uri2id"])

    return CLASS_AR



from dataclasses import dataclass
from io import StringIO
import orjson, csv
from typing import Callable, Dict, List, Mapping, Set, Tuple
from sand.config import SETTINGS
from sand.models.entity import NIL_ENTITY, Entity
from sand.models.ontology import OntProperty, OntPropertyAR, OntPropertyDataType
from sand.models.table import Table, TableRow
import sm.outputs as O
import sm.misc as M
from sm.namespaces.wikidata import WikidataNamespace
import drepr.models.sm as drepr_sm


prefixes = WikidataNamespace.create().prefix2ns.copy()
prefixes.update(drepr_sm.SemanticModel.get_default_prefixes())


# mapping from predefined datatypes to D-REPR datatype
datatype_mapping: Mapping[OntPropertyDataType, drepr_sm.DataType] = {
    "globe-coordinate": drepr_sm.DataType.geo_wktLiteral,
    "url": drepr_sm.DataType.xsd_anyURI,
    "external-id": drepr_sm.DataType.xsd_anyURI,
    "string": drepr_sm.DataType.xsd_string,
    "quantity": None,
    "time": drepr_sm.DataType.xsd_dateTime,
}


def get_drepr_sm(
    sm: O.SemanticModel,
    id2props: Mapping[str, OntProperty],
    get_attr_id: Callable[[int], str],
    get_ent_attr_id: Callable[[int], str],
) -> drepr_sm.SemanticModel:
    """Convert sm model into drepr model.

    Args:
        sm: the semantic model we want to convert
        id2props: mapping from the id to ontology property
        get_attr_id: get attribute id from column index
        get_ent_attr_id: for each entity column, to generate url, we create an extra attribute containing the entity uri, this function get its id based on the column index
    """
    nodes = {}
    edges = {}

    for node in sm.nodes():
        if isinstance(node, O.ClassNode):
            nodes[str(node.id)] = drepr_sm.ClassNode(
                node_id=str(node.id), label=node.rel_uri
            )
        elif isinstance(node, O.DataNode):
            # find data type of this node, when they have multiple data types
            # that do not agree with each other, we don't set the datatype
            # usually, that will be displayed from the UI so users know that

            datatypes: Set[OntPropertyDataType] = {
                id2props[OntProperty.uri2id(inedge.abs_uri)].datatype
                for inedge in sm.in_edges(node.id)
            }
            datatype = (
                datatype_mapping.get(list(datatypes)[0], None)
                if len(datatypes) == 1
                else None
            )

            nodes[str(node.id)] = drepr_sm.DataNode(
                node_id=str(node.id),
                attr_id=get_attr_id(node.col_index),
                data_type=datatype,
            )
        elif isinstance(node, O.LiteralNode):
            if node.datatype == O.LiteralNodeDataType.Entity:
                datatype = drepr_sm.DataType.xsd_anyURI
            else:
                assert node.datatype == O.LiteralNodeDataType.String
                datatype = drepr_sm.DataType.xsd_string

            nodes[str(node.id)] = drepr_sm.LiteralNode(
                node_id=str(node.id), value=node.value, data_type=datatype
            )

    used_ids = {x for edge in sm.edges() for x in [str(edge.source), str(edge.target)]}
    for node_id in set(nodes.keys()).difference(used_ids):
        del nodes[node_id]

    for edge in sm.edges():
        edges[len(edges)] = drepr_sm.Edge(
            edge_id=len(edges),
            source_id=str(edge.source),
            target_id=str(edge.target),
            label=edge.rel_uri,
        )

    # add drepr:uri relationship
    for node in get_entity_data_nodes(sm):
        new_node_id = str(node.id) + ":ents"
        nodes[new_node_id] = drepr_sm.DataNode(
            node_id=new_node_id,
            attr_id=get_ent_attr_id(node.col_index),
            data_type=drepr_sm.DataType.xsd_anyURI,
        )
        inedges = [
            inedge
            for inedge in sm.in_edges(node.id)
            if inedge.abs_uri in SETTINGS["semantic_model"]["identifiers"]
        ]
        assert len(inedges) == 1
        inedge = inedges[0]
        edges[len(edges)] = drepr_sm.Edge(
            edge_id=len(edges),
            source_id=str(inedge.source),
            target_id=new_node_id,
            # special predicate telling drepr to use as uri of entity, instead of generating a blank node
            label="drepr:uri",
        )

    return drepr_sm.SemanticModel(
        nodes=nodes,
        edges=edges,
        prefixes=prefixes,
    )


def get_entity_data_nodes(sm: O.SemanticModel) -> List[O.DataNode]:
    ident_props = SETTINGS["semantic_model"]["identifiers"]
    ent_dnodes = []
    for node in sm.iter_nodes():
        if not isinstance(node, O.DataNode):
            continue

        stypes = sm.get_semantic_types_of_column(node.col_index)
        if any(stype.predicate_abs_uri in ident_props for stype in stypes):
            ent_dnodes.append(node)

    return ent_dnodes



import os
from pathlib import Path

import click
from loguru import logger
from peewee import fn
from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop
from tornado.wsgi import WSGIContainer

from sand.config import SETTINGS
from sand.models import (
    Project,
    SemanticModel,
    Table,
    TableRow,
    db as dbconn,
    init_db,
)
from sand.commands.load import load_dataset


@click.command()
@click.option("-d", "--db", required=True, help="smc database file")
def init(db):
    """Init database"""
    init_db(db)
    dbconn.create_tables([Project, Table, TableRow, SemanticModel], safe=True)
    if Project.select().where(fn.Lower(Project.name) == "default").count() == 0:
        Project(name="Default", description="The default project").save()


@click.command()
@click.option("-d", "--db", required=True, help="smc database file")
@click.option(
    "--externaldb",
    default="",
    help="Folder contains external databases containing entities & ontologies",
)
@click.option(
    "--externaldb-proxy",
    is_flag=True,
    help="Enable proxy on the externaldb",
)
@click.option("--wsgi", is_flag=True, help="Whether to use wsgi server")
@click.option("-p", "--port", default=5524, help="Listening port")
@click.option(
    "--certfile", default=None, help="Path to the certificate signing request"
)
@click.option("--keyfile", default=None, help="Path to the key file")
def start(
    db: str,
    externaldb: str,
    externaldb_proxy: bool,
    wsgi: bool,
    port: int,
    certfile: str,
    keyfile: str,
):
    init_db(db)

    if externaldb.strip() != "":
        for cfg in [SETTINGS[k] for k in ["entity", "ont_classes", "ont_props"]]:
            cfg["args"]["dbfile"] = os.path.join(
                externaldb, Path(cfg["args"]["dbfile"]).name
            )
            cfg["args"]["proxy"] = externaldb_proxy

    if certfile is None or keyfile is None:
        ssl_options = None
    else:
        ssl_options = {"certfile": certfile, "keyfile": keyfile}
        assert not wsgi

    from sand.app import app

    if wsgi:
        app.run(host="0.0.0.0", port=port)
    else:
        logger.info("Start server in non-wsgi mode")
        http_server = HTTPServer(WSGIContainer(app), ssl_options=ssl_options)
        http_server.listen(port)
        IOLoop.instance().start()


@click.command()
@click.option("-d", "--db", required=True, help="smc database file")
@click.option("--description", required=True, help="Description of the project")
@click.argument("name")
def create(db, description: str, name: str):
    """Create project if not exist"""
    init_db(db)
    Project(name=name, description=description).save()


@click.group()
def cli():
    pass


cli.add_command(init)
cli.add_command(start)
cli.add_command(create)
cli.add_command(load_dataset)


if __name__ == "__main__":
    cli()



from dataclasses import dataclass
from functools import partial
import os
from re import L
from typing import Dict, List, Optional

from gena import generate_readonly_api_4dict, generate_api
from gena import generate_app
from gena.deserializer import (
    deserialize_dict,
    generate_deserializer,
    get_dataclass_deserializer,
    get_deserializer_from_type,
)
from grams.algorithm.inferences.features.tree_utils import TreeStruct
from grams.inputs.context import Attribute
from hugedict.chained_mapping import ChainedMapping
from sand.deserializer import deserialize_graph
from sand.models import SemanticModel, EntityAR, Project, Table, TableRow
from sand.models.entity import NIL_ENTITY, Entity
from sand.models.ontology import OntClass, OntClassAR, OntProperty, OntPropertyAR
from sand.models.table import CandidateEntity, Link
from sand.plugins.drepr.relational2rdf import relational2rdf
from sand.serializer import (
    get_label,
    serialize_class,
    serialize_entity,
    batch_serialize_sms,
    serialize_property,
)
import sm.outputs as O
from sand.plugins.wikidata import DEFAULT_ONT_CLASSES, DEFAULT_ONT_PROPS
from flask import json, jsonify, request, make_response
from peewee import Model as PeeweeModel, DoesNotExist, fn

from werkzeug.exceptions import BadRequest, NotFound

table_bp = generate_api(
    Table,
    deserializers=generate_deserializer(Table, {Attribute: deserialize_dict}),
)
table_row_bp = generate_api(TableRow)
deser_list_links = get_deserializer_from_type(List[Link], {})
assert deser_list_links is not None


@dataclass
class UpdateColumnLinksInput:
    table: int
    column: int
    text: str
    entity_id: Optional[str]


deser_update_column_links = get_dataclass_deserializer(UpdateColumnLinksInput, {})
assert deser_update_column_links is not None


@table_bp.route(f"/{table_bp.name}/<id>/export-models", methods=["GET"])
def export_sms(id: int):
    subquery = (
        SemanticModel.select(
            SemanticModel.id, fn.MAX(SemanticModel.version).alias("version")
        )
        .where(SemanticModel.table == id)
        .group_by(SemanticModel.table, SemanticModel.name)
        .alias("q1")
    )

    query = (
        SemanticModel.select()
        .where(SemanticModel.table == id)
        .join(subquery, on=(SemanticModel.id == subquery.c.id))
    )

    sms: List[O.SemanticModel] = [r.data for r in query]
    ontprops = OntPropertyAR()
    ontclasses = OntClassAR()
    uri2lbl = partial(get_label, ontprops=ontprops, ontclasses=ontclasses)

    for sm in sms:
        for n in sm.iter_nodes():
            if isinstance(n, O.ClassNode):
                if n.readable_label is None:
                    n.readable_label = uri2lbl(n.abs_uri, is_class=True) or n.rel_uri
        for e in sm.iter_edges():
            if e.readable_label is None:
                e.readable_label = uri2lbl(e.abs_uri, is_class=False) or e.rel_uri

    resp = jsonify([sm.to_dict() for sm in sms])
    if request.args.get("attachment", "false") == "true":
        resp.headers["Content-Disposition"] = "attachment; filename=export.json"
    return resp


@table_bp.route(f"/{table_bp.name}/<id>/export", methods=["GET"])
def export_data(id: int):
    # load table
    table: Table = Table.get_by_id(id)

    # load models
    subquery = (
        SemanticModel.select(
            SemanticModel.id, fn.MAX(SemanticModel.version).alias("version")
        )
        .where(SemanticModel.table == table)
        .group_by(SemanticModel.table, SemanticModel.name)
        .alias("q1")
    )
    query = (
        SemanticModel.select()
        .where(SemanticModel.table == table)
        .join(subquery, on=(SemanticModel.id == subquery.c.id))
    )
    sms: List[SemanticModel] = list(query)

    if len(sms) == 0:
        raise BadRequest("Exporting data requires the table to be modeled")

    sm_name = request.args["sm"] if "sm" in request.args else sms[0].name
    sms = [sm for sm in sms if sm.name == sm_name]
    if len(sms) == 0:
        raise BadRequest(f"The semantic model with name {sm_name} is not found")
    sm = sms[0]

    # load rows
    rows: List[TableRow] = list(TableRow.select().where(TableRow.table == table))

    # export the data using drepr library
    content = relational2rdf(table, rows, sm.data)
    resp = make_response(content)
    resp.headers["Content-Type"] = "text/ttl; charset=utf-8"
    if request.args.get("attachment", "false") == "true":
        resp.headers["Content-Disposition"] = "attachment; filename=export.ttl"
    return resp


@table_row_bp.route(f"/{table_row_bp.name}/<id>/cells/<column>", methods=["PUT"])
def update_cell_link(id: int, column: int):
    """
    Update the link entity & candidate entities for a given table row and column.
    """
    try:
        row: TableRow = TableRow.get_by_id(id)
    except DoesNotExist:
        raise NotFound(f"Record {id} does not exist")

    try:
        column = int(column)
        if column >= len(row.row):
            raise BadRequest(f"Invalid column {column} value")
    except ValueError:
        raise BadRequest(f"Column {column} is not an integer")

    request_json = request.get_json()
    if request_json is None:
        raise BadRequest("Missing request body")

    if "links" not in request_json:
        raise KeyError(f"Field 'links' is required")
    if not isinstance(request_json["links"], list) or not all(
        isinstance(link, dict) for link in request_json["links"]
    ):
        raise KeyError(f"Field 'links' must be a list of dictionaries")

    if (
        len(request_json["links"]) > 0
        and "candidate_entities" not in request_json["links"][0]
    ):
        # add back the candidate so we can deserialize them
        for link in request_json["links"]:
            link["candidate_entities"] = []
        links: List[Link] = deser_list_links(request_json["links"])

        # then, we just need to update the links' entities
        if len(row.links.get(str(column), [])) == 0:
            row.links[str(column)] = links
        else:
            if len(row.links[str(column)]) != len(links):
                raise BadRequest(
                    "Number of links in request does not match number of links in database"
                )
            for i, link in enumerate(links):
                db_link = row.links[str(column)][i]
                db_link.start = link.start
                db_link.end = link.end
                db_link.entity_id = link.entity_id
    else:
        row.links[str(column)] = deser_list_links(request_json["links"])

    row.save()
    return jsonify({"success": True})


@table_row_bp.route(f"/{table_row_bp.name}/update_column_links", methods=["PUT"])
def update_column_links():
    request_json = request.get_json()
    if request_json is None:
        raise BadRequest("Missing request body")

    args: UpdateColumnLinksInput = deser_update_column_links(request_json)
    table: Table = Table.get_by_id(args.table)
    if args.column >= len(table.columns):
        raise BadRequest(f"Invalid column {args.column} value")

    rows: List[TableRow] = list(TableRow.select().where(TableRow.table == table))
    str_column = str(args.column)
    for row in rows:
        if str(row.row[args.column]) != args.text:
            continue

        if str_column not in row.links:
            row.links[str_column] = [
                Link(
                    start=0,
                    end=len(str(row.row[args.column])),
                    url=None,
                    entity_id=args.entity_id,
                    candidate_entities=[],
                )
            ]
        else:
            db_link = row.links[str_column][0]
            db_link.start = 0
            db_link.end = len(str(row.row[args.column]))
            db_link.entity_id = args.entity_id

        row.save()
    return jsonify({"success": True})



from gena.deserializer import get_dataclass_deserializer
import sm.outputs as O
from sm.outputs.semantic_model import LiteralNodeDataType
from sand.plugins.wikidata import get_rel_uri


def deserialize_graph(value) -> O.SemanticModel:
    if not isinstance(value, dict):
        raise ValueError(f"expect dictionary but get {type(value)}")
    if "nodes" not in value:
        raise ValueError("missing `nodes`")
    if "edges" not in value:
        raise ValueError("missing `edges`")
    if not isinstance(value["nodes"], list):
        raise ValueError(f"expect `nodes` to be a list but get {type(value['nodes'])}")
    if not isinstance(value["edges"], list):
        raise ValueError(f"expect `edges` to be a list but get {type(value['edges'])}")

    sm = O.SemanticModel()
    idmap = {}

    for node in value["nodes"]:
        if not isinstance(node, dict):
            raise ValueError(f"expect node to be a dictionary but get {type(node)}")
        for k in ["id", "type"]:
            if k not in node:
                raise ValueError(f"missing `{k}` in node")
            if not isinstance(node[k], (int, str)):
                raise ValueError(
                    f"expect `{k}` to be a string or number but get {type(node[k])}"
                )
        # if isinstance(node["id"], str):
        #     if not node["id"].isdigit():
        #         raise ValueError(
        #             f"expect `id` to be a string number but get {type(node['id'])}"
        #         )
        #     node["id"] = int(node["id"])
        # if not isinstance(node["id"], int):
        #     raise ValueError(f"expect `id` to be a number but get {type(node['id'])}")

        if node["type"] == "class_node":
            for k in ["uri", "label"]:
                if k not in node:
                    raise ValueError(f"missing `{k}` in class_node")
                if not isinstance(node[k], str):
                    raise ValueError(
                        f"expect `{k}` to be a string but get {type(node[k])}"
                    )
            if "approximation" not in node:
                raise ValueError(f"missing `approximation` in class_node")
            if not isinstance(node["approximation"], bool):
                raise ValueError(
                    f"expect `approximation` to be a boolean but get {type(node['approximation'])}"
                )

            n = O.ClassNode(
                abs_uri=node["uri"],
                rel_uri=get_rel_uri(node["uri"]),
                approximation=node["approximation"],
                readable_label=node["label"],
            )
        elif node["type"] == "data_node":
            for k in ["label"]:
                if k not in node:
                    raise ValueError(f"missing `{k}` in data_node")
                if not isinstance(node[k], str):
                    raise ValueError(
                        f"expect `{k}` to be a string but get {type(node[k])}"
                    )
            if "column_index" not in node:
                raise ValueError(f"missing `column_index` in data_node")
            if not isinstance(node["column_index"], int):
                raise ValueError(
                    f"expect `column_index` to be an integer but get {type(node['column_index'])}"
                )

            n = O.DataNode(
                # id=node["id"],
                col_index=node["column_index"],
                label=node["label"],
            )
        elif node["type"] == "literal_node":
            for k in ["label"]:
                if k not in node:
                    raise ValueError(f"missing `{k}` in literal_node")
                if not isinstance(node[k], str):
                    raise ValueError(
                        f"expect `{k}` to be a string but get {type(node[k])}"
                    )
            if "is_in_context" not in node:
                raise ValueError(f"missing `is_in_context` in literal_node")
            if not isinstance(node["is_in_context"], bool):
                raise ValueError(
                    f"expect `is_in_context` to be a boolean but get {type(node['is_in_context'])}"
                )

            if "value" not in node:
                raise ValueError(f"missing `value` in literal_node")
            if not isinstance(node["value"], dict):
                raise ValueError(
                    f"expect `value` to be a dictionary but get {type(node['value'])}"
                )
            if "type" not in node["value"]:
                raise ValueError(f"missing `type` in literal_node")
            if not isinstance(node["value"]["type"], str):
                raise ValueError(
                    f"expect `type` to be a string but get {type(node['value']['type'])}"
                )
            datatype = LiteralNodeDataType(node["value"]["type"])

            if datatype == LiteralNodeDataType.String:
                if "value" not in node["value"]:
                    raise ValueError(f"missing `value` in literal_node")
                if not isinstance(node["value"]["value"], str):
                    raise ValueError(
                        f"expect `value` to be a string but get {type(node['value']['value'])}"
                    )
                node_value = node["value"]["value"]
            elif datatype == LiteralNodeDataType.Entity:
                if "uri" not in node["value"]:
                    raise ValueError(f"missing `uri` in literal_node")
                if not isinstance(node["value"]["uri"], str):
                    raise ValueError(
                        f"expect `uri` to be a string but get {type(node['value']['uri'])}"
                    )
                node_value = node["value"]["uri"]
            else:
                raise ValueError(f"unknown datatype {node['value']['type']}")

            n = O.LiteralNode(
                # id=node["id"],
                value=node_value,
                is_in_context=node["is_in_context"],
                readable_label=node["label"],
                datatype=datatype,
            )
        else:
            raise ValueError(f"unknown node type {node['type']}")

        idmap[node["id"]] = sm.add_node(n)

    for edge in value["edges"]:
        if not isinstance(edge, dict):
            raise ValueError(f"expect edge to be a dictionary but get {type(edge)}")

        for k in ["uri", "label"]:
            if k not in edge:
                raise ValueError(f"missing `{k}` in edge")
            if not isinstance(edge[k], str):
                raise ValueError(f"expect `{k}` to be a string but get {type(edge[k])}")

        for k in ["source", "target"]:
            if k not in edge:
                raise ValueError(f"missing `{k}` in edge")
            if not isinstance(edge[k], (str, int)):
                raise ValueError(
                    f"expect `{k}` to be string or number but get {type(edge[k])}"
                )
            if edge[k] not in idmap:
                raise KeyError(f"expect edge's {k} to exist")
            # if isinstance(edge[k], str):
            #     if not edge[k].isdigit():
            #         raise ValueError(
            #             f"expect `{k}` to be a string number but get {type(edge[k])}"
            #         )
            #     edge[k] = int(edge[k])
            # if not isinstance(edge[k], int):
            #     raise ValueError(f"expect `{k}` to be a number but get {type(edge[k])}")

        if "approximation" not in edge:
            raise ValueError(f"missing `approximation` in edge")
        if not isinstance(edge["approximation"], bool):
            raise ValueError(
                f"expect `approximation` to be a boolean but get {type(edge['approximation'])}"
            )

        e = O.Edge(
            source=idmap[edge["source"]],
            target=idmap[edge["target"]],
            abs_uri=edge["uri"],
            rel_uri=get_rel_uri(edge["uri"]),
            approximation=edge["approximation"],
            readable_label=edge["label"],
        )
        if not (sm.has_node(e.source) and sm.has_node(e.target)):
            raise ValueError(f"edge {str(edge)} connects to unknown nodes")
        sm.add_edge(e)
    return sm



import os

from pathlib import Path


_ROOT_DIR = Path(os.path.abspath(__file__)).parent.parent
PACKAGE_DIR = str(Path(os.path.abspath(__file__)).parent)
FROM_SITEPACKAGES = _ROOT_DIR.name == "site-packages"

CACHE_SIZE = 10240

SETTINGS = {
    "entity": {
        "constructor": "sand.plugins.wikidata.get_entity_db",
        "uri2id": "sm.namespaces.WikidataNamespace.get_entity_id",
        "id2uri": "sm.namespaces.WikidataNamespace.get_entity_abs_uri",
        "args": {
            "dbfile": "/tmp/wdentities.db",
            "proxy": True,
        },
        # extra entities
        "default": "sand.models.entity.DEFAULT_ENTITY",
        # mapping from entity's namespace to the property id that will be used to indicate `instance_of` relationship
        "instanceof": {
            "http://www.wikidata.org": "P31",
        },
        # id of an nil entity
        "nil": "drepr:nil",
        # template for new entity uri
        "new_entity_template": "http://www.wikidata.org/entity/{id}",
    },
    "ont_classes": {
        "constructor": "sand.plugins.wikidata.get_ontclass_db",
        "uri2id": "sand.plugins.wikidata.uri2id",
        "args": {
            "dbfile": "/tmp/wdclasses.db",
            "proxy": True,
        },
        # extra classes
        "default": "sand.plugins.wikidata.WD_ONT_CLASSES",
    },
    "ont_props": {
        "constructor": "sand.plugins.wikidata.get_ontprop_db",
        "uri2id": "sand.plugins.wikidata.uri2id",
        "args": {
            "dbfile": "/tmp/wdprops.db",
            "proxy": True,
        },
        # extra props
        "default": "sand.models.ontology.DEFAULT_ONT_PROPS",
    },
    "semantic_model": {
        # list of properties' uris that when a column is tagged with one of them, the column is an entity column
        "identifiers": [
            "http://www.w3.org/2000/01/rdf-schema#label",
        ],
        # list of uri of classes that are used as intermediate nodes to represent n-ary relationships, e.g., wikidata's statement
        "statements": ["http://wikiba.se/ontology#Statement"],
    },
    "assistants": {
        # list of assistants' names and their models
        # "grams": "sand.plugins.grams.GRAMSAssistant",
        "mtab": "sand.plugins.mtab.MTabAssistant",
        # "default": "mtab",
    },
}



from typing import List, Tuple
from grams.inputs.linked_table import LinkedTable
from sand.models import (
    Project,
    Value,
    Link,
    Table,
    ContextPage,
    TableRow,
    SemanticModel,
)


def convert_linked_table(tbl: LinkedTable) -> Tuple[Table, List[TableRow]]:
    """Convert linked table to smc's tables and rows.
    Note that the project of the table is not set, which is supposed to be set later.
    """
    if tbl.context.page_entity_id is None:
        context_values = []
    else:
        context_values = [Value("entityid", tbl.context.page_entity_id)]

    if tbl.context.page_url is not None:
        context_page = ContextPage(
            tbl.context.page_url,
            tbl.context.page_title or "",
            tbl.context.page_entity_id,
        )
    else:
        context_page = None

    links = {}
    for ri, rlink in enumerate(tbl.links):
        if ri not in links:
            links[ri] = {}
        for ci, clinks in enumerate(rlink):
            ci = str(ci)
            if len(clinks) > 0:
                links[ri][ci] = []
                for clink in clinks:
                    links[ri][ci].append(
                        Link(
                            clink.start,
                            clink.end,
                            clink.url,
                            clink.entity_id,
                            [],
                        )
                    )

    columns = [col.name for col in tbl.table.columns]
    mtbl = Table(
        name=tbl.id,
        description="",
        columns=columns,
        size=tbl.size(),
        context_values=context_values,
        context_tree=[],
        context_page=context_page,
        project=None,
    )
    mrows = []

    for ri in range(tbl.size()):
        row = [tbl.table[ri, ci] for ci in range(len(columns))]
        mrows.append(TableRow(table=mtbl, index=ri, row=row, links=links[ri]))

    return mtbl, mrows



import copy
from rdflib import RDFS
import subprocess
from abc import ABC, abstractmethod
from pathlib import Path
import threading
from typing import Dict, List, Tuple
from flask.blueprints import Blueprint
from sm.misc.funcs import V, import_func
from sm.misc.deser import (
    deserialize_csv,
    deserialize_json,
    deserialize_text,
    serialize_csv,
)
from sm.namespaces.wikidata import WikidataNamespace
import sm.outputs as O
from sand.controllers.assistant import Assistant
from sand.models.base import init_db, db

from sand.models.table import CandidateEntity, Link, Table, TableRow
from flask import request, jsonify
from uuid import uuid4


class MTabAssistant(Assistant):
    def __init__(self):
        self.cache_dir = Path(f"/tmp/mtab")
        self.cache_dir.mkdir(exist_ok=True)

        self.id2label = {}

    def predict(self, table: Table, rows: List[TableRow]):
        content = [table.columns]
        for row in rows:
            content.append([str(x) for x in row.row])

        cwd = self.cache_dir / f"p{table.project.id}-t:{table.id}-{table.name}"
        cwd.mkdir(exist_ok=True)
        infile = cwd / "infile.csv"
        outfile = cwd / "outfile.json"

        rerun = True
        if infile.exists() and outfile.exists():
            serialize_csv(content, str(infile) + ".tmp")
            if deserialize_text(infile) == deserialize_text(str(infile) + ".tmp"):
                rerun = False

        if rerun:
            serialize_csv(content, infile)
            subprocess.check_call(
                [
                    "curl",
                    "-X",
                    "POST",
                    "-F",
                    f"file=@{str(infile)}",
                    "https://mtab.app/api/v1/mtab?limit=1000",
                    "-o",
                    outfile,
                ]
            )

        record = deserialize_json(outfile)
        assert record["status"] == "Success"

        semantic = record["tables"][0]["semantic"]

        cpa, cta = [], {}
        for item in semantic["cpa"]:
            source, target = item["target"]
            # assert len(item["annotation"]) == 1
            uri = item["annotation"][0]["wikidata"]
            label = item["annotation"][0]["label"]
            assert uri.startswith("http://www.wikidata.org/prop/direct/")
            pid = uri.replace("http://www.wikidata.org/prop/direct/", "")
            self.id2label[pid] = label
            cpa.append((source, target, pid))

        for item in semantic["cta"]:
            # assert len(item["annotation"]) == 1
            uri = item["annotation"][0]["wikidata"]
            label = item["annotation"][0]["label"]
            assert uri.startswith("http://www.wikidata.org/entity/")
            qid = uri.replace("http://www.wikidata.org/entity/", "")
            cta[item["target"]] = qid
            self.id2label[qid] = label

        pred_sm = self.create_sm_from_cta_cpa(table.columns, cpa, cta)

        # clean previous data -- temporary solution for now before we tag each result with
        # individual method
        rows = copy.deepcopy(rows)
        for row in rows:
            row.links = {}

        for ann in semantic["cea"]:
            ri, ci = ann["target"]
            ri = ri - 1
            if ri == -1:
                continue

            uri = ann["annotation"]["wikidata"]
            if uri.startswith("http://www.wikidata.org/entity/"):
                entid = uri.replace("http://www.wikidata.org/entity/", "")
            elif uri.startswith("http://www.wikidata.org/prop/direct/"):
                entid = uri.replace("http://www.wikidata.org/prop/direct/", "")
            else:
                raise ValueError(uri)

            if str(ci) not in rows[ri].links:
                rows[ri].links[str(ci)] = [
                    Link(
                        start=0,
                        end=len(str(rows[ri].row[ci])),
                        url=None,
                        entity_id=None,
                        candidate_entities=[],
                    )
                ]
            rows[ri].links[str(ci)][0].candidate_entities.append(
                CandidateEntity(entity_id=entid, probability=1.0)
            )

        return pred_sm, rows

    def create_sm_from_cta_cpa(
        self,
        columns: List[str],
        rels: List[Tuple[int, int, str]],
        types: Dict[int, str],
    ) -> O.SemanticModel:
        """Create a semantic model from a list of columns, CPA, and CTA.

        Args:
            columns: list of column names
            rels: list of (source column index, target column index, prop id not url)
            types: a mapping from column index to WDEntity id (not url)
        """
        sm = O.SemanticModel()

        # create class nodes first
        classmap = {}

        for cid, qnode_id in types.items():
            dnode = O.DataNode(
                col_index=cid,
                label=columns[cid],
            )

            # somehow, they may end-up predict multiple classes, we need to select one
            if qnode_id.find(" ") != -1:
                qnode_id = qnode_id.split(" ")[0]
            curl = WikidataNamespace.get_entity_abs_uri(qnode_id)

            try:
                cnode_label = f"{self.id2label[qnode_id]} ({qnode_id})"
            except KeyError:
                cnode_label = WikidataNamespace.get_entity_rel_uri(qnode_id)
            cnode = O.ClassNode(
                abs_uri=curl,
                rel_uri=WikidataNamespace.get_entity_rel_uri(qnode_id),
                readable_label=cnode_label,
            )
            sm.add_node(dnode)
            sm.add_node(cnode)
            sm.add_edge(
                O.Edge(
                    source=cnode.id,
                    target=dnode.id,
                    abs_uri=str(RDFS.label),
                    rel_uri="rdfs:label",
                    readable_label="rdfs:label",
                )
            )

            classmap[dnode.id] = cnode.id

        # now complete the relationships
        for source_cid, target_cid, prop in rels:
            if sm.has_data_node(source_cid):
                source = sm.get_data_node(source_cid)
                # if source node is an NE column, we need to use a class id instead
                if source.id in classmap:
                    source = sm.get_node(classmap[source.id])
            else:
                # create a data node
                source = O.DataNode(
                    col_index=source_cid,
                    label=columns[source_cid],
                )
                sm.add_node(source)

            if sm.has_data_node(target_cid):
                # if target node is an NE column, we need to use a class id instead
                target = sm.get_data_node(target_cid)
                if target.id in classmap:
                    target = sm.get_node(classmap[target.id])
            else:
                # create a data node
                target = O.DataNode(col_index=target_cid, label=columns[target_cid])
                sm.add_node(target)

            sm.add_edge(
                O.Edge(
                    source=source.id,
                    target=target.id,
                    abs_uri=WikidataNamespace.get_prop_abs_uri(prop),
                    rel_uri=WikidataNamespace.get_prop_rel_uri(prop),
                    readable_label=f"{self.id2label[prop]} ({prop})",
                )
            )

        # complete missing data nodes
        for ci, col in enumerate(columns):
            if sm.has_data_node(ci):
                continue
            sm.add_node(O.DataNode(col_index=ci, label=col))
        return sm


if __name__ == "__main__":
    init_db("/Users/rook/workspace/sm-dev/data/home/databases/smc.db")
    table = Table.get_by_id(501)
    rows = list(
        TableRow.select().where(TableRow.table == table).order_by(TableRow.index)
    )

    assistant = MTabAssistant()
    resp = assistant.predict(table, rows)
    print(resp)



import copy
import csv
from dataclasses import asdict
import os
from functools import partial
from io import StringIO
from typing import List, Optional, Union, cast, no_type_check_decorator

import orjson
import sm.misc as M
import sm.outputs as O
from flask import json, jsonify, request
from gena import generate_api, generate_app, generate_readonly_api_4dict
from gena.deserializer import (
    deserialize_dict,
    generate_deserializer,
    get_dataclass_deserializer,
)
from grams.inputs.context import Attribute
from hugedict.chained_mapping import ChainedMapping
from peewee import DoesNotExist
from peewee import Model as PeeweeModel
from peewee import fn
from sand.controllers.helpers.upload import (
    ALLOWED_EXTENSIONS,
    CSVParserOpts,
    RawTable,
    UploadingTable,
    get_extension,
    parse_upload,
    save_upload,
)
from sand.deserializer import deserialize_graph
from sand.models import (
    EntityAR,
    Project,
    SemanticModel,
    Table,
    TableRow,
)
from sand.models.entity import Value
from sand.models.ontology import OntClass, OntClassAR, OntProperty, OntPropertyAR
from sand.plugins.wikidata import DEFAULT_ONT_CLASSES, DEFAULT_ONT_PROPS
from sand.serializer import (
    batch_serialize_sms,
    get_label,
    serialize_class,
    serialize_entity,
    serialize_property,
)
from werkzeug.datastructures import FileStorage
from werkzeug.exceptions import BadRequest
from sand.controllers.helpers.upload import CSVParserOpts, JSONParserOpts

project_bp = generate_api(Project)

deser_CSVParserOpts = get_dataclass_deserializer(CSVParserOpts, {})
deser_JSONParserOpts = get_dataclass_deserializer(JSONParserOpts, {})
assert deser_CSVParserOpts is not None
assert deser_JSONParserOpts is not None


@project_bp.route(f"/{project_bp.name}/<id>/upload", methods=["POST"])
def upload(id: int):
    """Upload tables to the project"""
    files = {}

    for file_id, file in request.files.items():
        if (
            file
            and file.filename is not None
            and get_extension(file.filename) in ALLOWED_EXTENSIONS
        ):
            files[file_id] = file

    if len(files) == 0:
        raise BadRequest(f"No files provided")

    if "parser_opts" in request.form:
        try:
            raw_parser_opts = orjson.loads(request.form["parser_opts"])
        except ValueError:
            raise BadRequest("Invalid vaue for `parser_opts`")

        if not isinstance(raw_parser_opts, dict):
            raise BadRequest(
                "Invalid value for `parser_opts`. Expect a dictionary of parser options for each file id"
            )

        if len(set(raw_parser_opts.keys()).difference(files.keys())) > 0:
            raise BadRequest(
                "Invalid value for `parser_opts`. Contains unknown file ids"
            )

        parser_opts = {}
        for file_id, opts in raw_parser_opts.items():
            if not isinstance(opts, dict) or "format" not in opts:
                raise BadRequest(
                    f"Invalid parser options for `{file_id}`. Expect a dictionary with a `format` key"
                )
            if opts["format"] == "csv":
                parser_opts[file_id] = deser_CSVParserOpts(opts)
            elif opts["format"] == "json":
                parser_opts[file_id] = deser_JSONParserOpts(opts)
            else:
                raise BadRequest(f"Invalid format `{file_id}`.")
    else:
        parser_opts = {}

    # parse the content
    tables: List[UploadingTable] = []
    for file_id, file in files.items():
        tables += parse_upload(parser_opts.get(file_id, None), file)

    if "selected_tables" in request.form:
        # signal that we go ahead with the selected tables and save it to the database
        try:
            selected_tables = orjson.loads(request.form["selected_tables"])
        except ValueError:
            raise BadRequest("Invalid value for `selected_tables`")

        raw_tables = [raw_table for table in tables for raw_table in table.tables]
        if (
            not isinstance(selected_tables, list)
            or len(selected_tables) == 0
            or any(
                not isinstance(x, int) or x < 0 or x >= len(raw_tables)
                for x in selected_tables
            )
        ):
            raise BadRequest("`selected_tables` must be a list of numbers")

        try:
            project = Project.get_by_id(id)
        except:
            raise BadRequest("Project not found")

        # rename duplicated tables
        names = {}
        for i, tbl in enumerate(raw_tables):
            if tbl.name not in names:
                names[tbl.name] = [i, 1]
            else:
                names[tbl.name][1] += 1
                raw_tables[names[tbl.name][0]].name = tbl.name + "-1"
                tbl.name = tbl.name + "-" + names[tbl.name][1]

        dbtables = save_upload(project, [raw_tables[i] for i in selected_tables])
        return jsonify(
            {"status": "success", "table_ids": [table.id for table in dbtables]}
        )

    return jsonify(
        {
            "status": "success",
            "tables": [
                {
                    "parser_opts": asdict(table.parser_opts),
                    "tables": [asdict(raw_table) for raw_table in table.tables],
                }
                for table in tables
            ],
        }
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, Dict, List, Literal, Union
from kgdata.wikidata.models.wdvalue import (
    ValueGlobeCoordinate,
    ValueMonolingualText,
    ValueQuantity,
    ValueTime,
)
from kgdata.wikidata.models.multilingual import (
    MultiLingualString,
    MultiLingualStringList,
)

from sand.config import SETTINGS
from sm.misc.funcs import import_func


# represent that there is no entity
NIL_ENTITY = SETTINGS["entity"]["nil"]


@dataclass
class Entity:
    id: str
    label: MultiLingualString
    aliases: MultiLingualStringList
    description: MultiLingualString
    properties: Dict[str, List[Statement]]

    @property
    def readable_label(self):
        return self.label

    @property
    def instanceof(self):
        """Get the property representing the instanceof relation."""
        raise NotImplementedError()

    @staticmethod
    def uri2id(uri: str) -> str:
        """Convert entity URI to entity ID."""
        raise NotImplementedError(
            "The method is set when its store is initialized. Check the call order to ensure `EntityAR` is called first"
        )

    @staticmethod
    def id2uri(id: str) -> str:
        """Convert entity ID to entity URI."""
        raise NotImplementedError(
            "The method is set when its store is initialized. Check the call order to ensure `EntityAR` is called first"
        )


@dataclass
class Statement:
    __slots__ = ("value", "qualifiers", "qualifiers_order")
    value: Value
    qualifiers: Dict[str, List[Value]]
    # list of qualifiers id that records the order (as dict lacks of order)
    qualifiers_order: List[str]


@dataclass
class Value:
    __slots__ = ("type", "value")
    type: Literal[
        "string", "time", "quantity", "globecoordinate", "entityid", "monolingualtext"
    ]
    value: Union[
        str,
        ValueGlobeCoordinate,
        ValueQuantity,
        ValueTime,
        ValueMonolingualText,
    ]

    def is_entity_id(self):
        return self.type == "entityid"


ENTITY_AR = None
DEFAULT_ENTITY = {
    NIL_ENTITY: Entity(
        id=NIL_ENTITY,
        label=MultiLingualString.en("NIL"),
        aliases=MultiLingualStringList(lang2values={"en": []}, lang="en"),
        description=MultiLingualString.en("the correct entity is absent in KG"),
        properties={},
    )
}


def check_nil(fn: Callable[[str], str]):
    def wrapper(x: str) -> str:
        return fn(x) if x != NIL_ENTITY else NIL_ENTITY

    return wrapper


def EntityAR() -> Dict[str, Entity]:
    global ENTITY_AR

    if ENTITY_AR is None:
        cfg = SETTINGS["entity"]
        func = import_func(cfg["constructor"])
        ENTITY_AR = func(**cfg["args"])
        Entity.uri2id = check_nil(import_func(cfg["uri2id"]))
        Entity.id2uri = check_nil(import_func(cfg["id2uri"]))

    return ENTITY_AR



import glob
import os
from pathlib import Path

import click
from tqdm.auto import tqdm
from loguru import logger
from sand.models import (
    Project,
    SemanticModel,
    db as dbconn,
    init_db,
)
from sand.plugins.grams_plugin import convert_linked_table
from sm.dataset import Example, load as sm_load
from grams.inputs import LinkedTable


@click.command(name="load")
@click.option("-d", "--db", required=True, help="smc database file")
@click.option("-p", "--project", default="default", help="Project name")
@click.option("--dataset", required=True, help="Path to tables")
@click.option(
    "-n",
    "--n-tables",
    default=-1,
    type=int,
    help="Number of tables to load (negative number or zero to load all)",
)
def load_dataset(db: str, project: str, dataset: str, n_tables: int):
    """Load a dataset into a project"""
    init_db(db)

    examples = sm_load(dataset, table_deser=LinkedTable.from_dict)
    if n_tables > 0:
        examples = examples[:n_tables]

    with dbconn:
        p = Project.get(name=project)
        for e in tqdm(examples, desc="Loading examples"):
            save_example(p, e)


def save_example(project: Project, example: Example[LinkedTable]):
    mtbl, mrows = convert_linked_table(example.table)
    mtbl.project = project
    mtbl.save()

    for row in mrows:
        row.save()

    for i, sm in enumerate(example.sms):
        SemanticModel(
            table=mtbl,
            name=f"sm-{i}",
            description="",
            version=1,
            data=sm,
        ).save()



from flask import Blueprint
from flask.json import jsonify
from sand.config import SETTINGS

setting_bp = Blueprint("settings", "settings")


@setting_bp.route(f"/{setting_bp.name}", methods=["GET"])
def get_settings():
    return jsonify(
        {
            "entity": {
                "nil": SETTINGS["entity"]["nil"],
                "instanceof": SETTINGS["entity"]["instanceof"],
            },
            "semantic_model": {
                "identifiers": SETTINGS["semantic_model"]["identifiers"],
                "statements": SETTINGS["semantic_model"]["statements"],
            },
        }
    )



from abc import ABC, abstractmethod
from dataclasses import dataclass
import threading
from typing import Dict, List, Optional, Tuple
from flask.blueprints import Blueprint
from gena.deserializer import (
    generate_deserializer,
    get_dataclass_deserializer,
)
from grams.algorithm.inferences.features.tree_utils import TreeStruct
from sm.misc.funcs import import_func
from sm.outputs.semantic_model import SemanticModel
from sand.config import SETTINGS
from functools import partial
from sand.models.table import Link, Table, TableRow
from flask import request, jsonify
from sand.serializer import batch_serialize_sms, get_label, serialize_graph
from sand.models.entity import NIL_ENTITY, Entity, EntityAR
from sand.models.ontology import OntProperty, OntClass, OntPropertyAR, OntClassAR
from operator import attrgetter

from peewee import Model as PeeweeModel, DoesNotExist, fn

from werkzeug.exceptions import BadRequest, NotFound


class Assistant(ABC):
    @abstractmethod
    def predict(
        self, table: Table, rows: List[TableRow]
    ) -> Tuple[Optional[SemanticModel], Optional[List[TableRow]]]:
        """Predict semantic model and link entities"""
        pass


assistant_bp = Blueprint("assistant", "assistant")


GetAssistantCache = threading.local()


def get_assistants() -> Dict[str, Assistant]:
    """Get all assistants"""
    global GetAssistantCache

    if not hasattr(GetAssistantCache, "assistants"):
        GetAssistantCache.assistants = {}
        for name, constructor in SETTINGS["assistants"].items():
            GetAssistantCache.assistants[name] = import_func(constructor)()

    return GetAssistantCache.assistants


@assistant_bp.route(f"/{assistant_bp.name}/predict/<table_id>", methods=["GET"])
def predict(table_id: int):

    table = Table.get_by_id(table_id)
    rows: List[TableRow] = list(
        TableRow.select().where(TableRow.table == table).order_by(TableRow.index)
    )

    assistants = get_assistants()
    selected_assistants = {
        name: assistants[name]
        for name in request.args.get("algorithm", "").split(",")
        if name in assistants
    }
    if len(selected_assistants) == 0:
        selected_assistants = assistants

    ontprops = OntPropertyAR()
    ontclasses = OntClassAR()
    uri2lbl = partial(get_label, ontprops=ontprops, ontclasses=ontclasses)

    outputs = {}
    for name, assistant in selected_assistants.items():
        sm, outputrows = assistant.predict(table, rows)
        if sm is not None:
            sm = serialize_graph(sm, uri2lbl, columns=None)
        if outputrows is not None:
            # preserved the manual entity linking from the users
            for outputrow, row in zip(outputrows, rows):
                assert outputrow.id == row.id
                for ci, links in outputrow.links.items():
                    if ci in row.links:
                        glinks: Dict[int, List[Link]] = {}
                        for link in row.links[ci]:
                            if link.entity_id is None:
                                continue
                            glinks.setdefault(link.start, []).append(link)
                        for link in links:
                            for glink in glinks.get(link.start, []):
                                if link.end == glink.end:
                                    link.entity_id = glink.entity_id
            outputrows = [row.to_dict() for row in outputrows]

        outputs[name] = {"sm": sm, "rows": outputrows}

    return jsonify(outputs)


@dataclass
class GatherColumnTypesInput:
    table: int
    column: int


deser_gather_column_types = get_dataclass_deserializer(GatherColumnTypesInput, {})
assert deser_gather_column_types is not None


@assistant_bp.route(f"/{assistant_bp.name}/column-types", methods=["POST"])
def gather_column_types():
    """Gather types of the column via its entities"""
    request_json = request.get_json()
    if request_json is None:
        raise BadRequest("Missing request body")

    args: GatherColumnTypesInput = deser_gather_column_types(request_json)

    try:
        table: Table = Table.get_by_id(args.table)
    except DoesNotExist:
        raise NotFound(f"Record {id} does not exist")

    if args.column >= len(table.columns):
        raise BadRequest(f"Invalid column {args.column} value")

    rows: List[TableRow] = list(TableRow.select().where(TableRow.table == table))
    column = str(args.column)
    ents: Dict[str, Optional[Entity]] = {}
    entities = EntityAR()
    ont_classes = OntClassAR()

    for row in rows:
        for link in row.links.get(column, []):
            if (
                link.entity_id is not None
                and link.entity_id != NIL_ENTITY
                and link.entity_id not in ents
            ):
                ents[link.entity_id] = entities.get(link.entity_id, None)
            for can_ent in link.candidate_entities:
                ents[can_ent.entity_id] = entities.get(can_ent.entity_id, None)

    classes: Dict[str, OntClass] = {}
    for ent in ents.values():
        if ent is None:
            continue
        for stmt in ent.properties.get(ent.instanceof, []):
            if stmt.value.is_entity_id():
                klass_id = stmt.value.value
                assert isinstance(klass_id, str)
                klass = ont_classes.get(klass_id, None)
                if klass is not None:
                    classes[klass_id] = klass

    tree = TreeStruct.construct(classes, get_parents=lambda id: classes[id].parents)
    return jsonify({p: list(cs) for p, cs in tree.p2cs.items()})



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sand',
 'sand.commands',
 'sand.controllers',
 'sand.controllers.helpers',
 'sand.models',
 'sand.plugins',
 'sand.plugins.drepr',
 'sand.plugins.drepr.raw_transformations']

package_data = \
{'': ['*'],
 'sand': ['www/*', 'www/static/css/*', 'www/static/js/*', 'www/static/media/*']}

install_requires = \
['Flask>=2.0.2,<3.0.0',
 'drepr==2.10.0-dev3',
 'gena>=1.1.2,<2.0.0',
 'kgdata>=2.3.0,<3.0.0',
 'lat_lon_parser>=1.3.0,<2.0.0',
 'loguru>=0.6.0',
 'orjson>=3.6.8,<4.0.0',
 'peewee>=3.14.4,<4.0.0',
 'python-dotenv>=0.19.0',
 'sem-desc>=3.5.2,<4.0.0',
 'sm-grams>=2.1.1,<3.0.0',
 'tornado>=6.1,<7.0']

entry_points = \
{'console_scripts': ['sand = sand.__main__:cli']}

setup_kwargs = {
    'name': 'web-sand',
    'version': '2.1.4',
    'description': 'UI for browsing/editing semantic descriptions',
    'long_description': '<h1 align="center">SAND</h1>\n\n<div align="center">\n\n![PyPI](https://img.shields.io/pypi/v/web-sand)\n![Python](https://img.shields.io/badge/python-v3.8+-blue.svg)\n[![GitHub Issues](https://img.shields.io/github/issues/usc-isi-i2/sand.svg)](https://github.com/usc-isi-i2/sand/issues)\n![Contributions welcome](https://img.shields.io/badge/contributions-welcome-orange.svg)\n[![License](https://img.shields.io/badge/license-MIT-blue.svg)](https://opensource.org/licenses/MIT)\n\n</div>\n\n## Table of Contents\n\n- [Introduction](#introduction)\n- [Installation](#installation)\n\n## Introduction\n\nSAND is an application to annotate semantic descriptions of tables and (optionally) linked records in tables to a target knowledge graph, then it can automatically export the table data to RDF, JSON-LD, etc. It also does basic data cleaning automatically based on the annotated semantic descriptions. SAND is designed to be customizable: you can plug in a new semantic modeling algorithm (which generates a semantic description automatically) or different knowledge graphs as long as you have a suitable plugin implemented SAND\'s plugin interface.\n\nMoreover, SAND offers an internal KG browsing and table filtering so you can interactively browsing and modeling your tables.\n\nFor a demo, please see: our [demo paper](./docs/paper.pdf), [demo video](https://github.com/usc-isi-i2/sand/wiki/Demo).\n\n<!-- For more documentation, please see [not available yet](). -->\n\n## Installation\n\nInstall from pip: `pip install -U web-sand`\n\n## Usage\n\n1. Start the webserver: `sand start -d <dbfile> --externaldb <folder_of_ent_and_ont_db>`\n2. Open the URL: `http://localhost:5524`\n\n## Development\n\n1. Install `yarn` and [`yalc`](https://github.com/wclr/yalc)\n2. Install dependencies: `yarn install`\n3. Start development server: `yarn start`\n4. Build production files: `yarn build`\n5. Build library files and publish to private index: `yarn build:lib && yalc public --private`\n',
    'author': 'Binh Vu',
    'author_email': 'binh@toan2.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/usc-isi-i2/sand',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import orjson
import sm.outputs as O
from peewee import CharField, ForeignKeyField, TextField, IntegerField
from playhouse.shortcuts import model_to_dict
from sand.models.base import BaseModel, BlobField
from sand.models.project import Project
from sand.models.table import Table


def ser_sm(pyvalue: O.SemanticModel):
    return orjson.dumps(pyvalue.to_dict())


def deser_sm(dbvalue: bytes):
    return O.SemanticModel.from_dict(orjson.loads(dbvalue))


class SemanticModel(BaseModel):
    table = ForeignKeyField(Table, backref="semantic_models", on_delete="CASCADE")
    # name of the semantic model as we may have more than one version for the same table
    name = CharField()
    description = TextField()  # description of this model
    version = IntegerField()  # version of the semantic model
    # the semantic model at this version
    data: O.SemanticModel = BlobField(serialize=ser_sm, deserialize=deser_sm)  # type: ignore

    class Meta:
        indexes = ((("table", "name"), True),)



from typing import Union
from gena.client import Client as IndividualClient
from grams.inputs.linked_table import LinkedTable
from pathlib import Path
from peewee import fn
from sand.models import init_db, SemanticModel, Project, Table
from sm.prelude import M
from tqdm import tqdm


class Client:
    def __init__(
        self,
        url: str = "http://localhost",
    ) -> None:
        self.url = url
        self.projects = IndividualClient(f"{url}/api/project")
        self.tables = IndividualClient(f"{url}/api/table")
        self.table_rows = IndividualClient(f"{url}/api/tablerow")
        self.semantic_models = IndividualClient(f"{url}/api/semanticmodel")

    def export(self, project_name: str, outdir: Union[str, Path]):
        outdir = Path(outdir)
        project = Project.get(name=project_name)
        tables = list(project.tables)

        for tbl in tqdm(tables):
            dname = LinkedTable._get_friendly_fs_id(tbl.name)
            subquery = (
                SemanticModel.select(
                    SemanticModel.id, fn.MAX(SemanticModel.version).alias("version")
                )
                .where(SemanticModel.table == tbl.id)
                .group_by(SemanticModel.table, SemanticModel.name)
                .alias("q1")
            )

            query = (
                SemanticModel.select()
                .where(SemanticModel.table == tbl.id)
                .join(subquery, on=(SemanticModel.id == subquery.c.id))
            )

            sms = list(query)
            outfile = outdir / dname / f"version.01.json"
            outfile.parent.mkdir(exist_ok=True, parents=True)
            M.serialize_json([sm.data.to_dict() for sm in sms], outfile, indent=2)

        return None



from dataclasses import dataclass
from io import StringIO
import orjson, csv
from typing import Dict, List, Set, Tuple
from sand.config import SETTINGS
from sand.models.entity import NIL_ENTITY, Entity
from sand.models.ontology import OntProperty, OntPropertyAR
from sand.models.table import Table, TableRow
import sm.outputs as O
import sm.misc as M
from drepr.models import (
    ResourceDataString,
    ResourceData,
)
from uuid import uuid4


def get_table_resource(table: Table, rows: List[TableRow]) -> ResourceData:
    """Return a CSV resource, in which the first row is the header"""
    f = StringIO()
    writer = csv.writer(
        f, delimiter=",", quoting=csv.QUOTE_MINIMAL, lineterminator="\n"
    )

    writer.writerow(table.columns)
    for row in rows:
        writer.writerow(row.row)

    return ResourceDataString(f.getvalue())


def get_entity_resource(
    table: Table, rows: List[TableRow], ent_columns: Set[int]
) -> ResourceData:
    """Return a CSV resource of matrix mapping each position in a table to a corresponding entity uri."""
    new_entity_template: str = SETTINGS["entity"]["new_entity_template"]
    ent_rows: List[List[str]] = []

    for ri, row in enumerate(rows):
        ent_rows.append([])
        for ci in range(len(row.row)):
            if ci not in ent_columns:
                ent_rows[-1].append("")
                continue

            links = row.links.get(str(ci), [])
            # TODO: what happens when there are multiple entities?
            # for now, just return the first entity
            ent = None
            for link in links:
                if link.entity_id is not None and link.entity_id != NIL_ENTITY:
                    ent = Entity.id2uri(link.entity_id)
                    break
            else:
                # generate new entity
                ent = new_entity_template.format(id=str(uuid4()))

            ent_rows[-1].append(ent)

    f = StringIO()
    writer = csv.writer(
        f, delimiter=",", quoting=csv.QUOTE_MINIMAL, lineterminator="\n"
    )

    writer.writerow(table.columns)
    for row in ent_rows:
        writer.writerow(row)
    return ResourceDataString(f.getvalue())



from dataclasses import dataclass
from io import StringIO
import orjson, csv
from typing import Dict, List, Set, Tuple, cast
from sand.config import SETTINGS
from sand.models.entity import NIL_ENTITY, Entity
from sand.models.ontology import OntProperty, OntPropertyAR, OntPropertyDataType
from sand.models.table import Table, TableRow
import sm.outputs as O
import sm.misc as M
from hugedict.chained_mapping import ChainedMapping
from drepr.models import (
    DRepr,
    Attr,
    Resource,
    ResourceType,
    CSVProp,
    Path,
    IndexExpr,
    RangeExpr,
    RangeAlignment,
    ValueAlignment,
    AlignmentType,
    AlignedStep,
    Preprocessing,
    PreprocessingType,
    PMap,
    ResourceDataString,
)
import drepr.models.sm as drepr_sm
from drepr.engine import execute, FileOutput, OutputFormat, MemoryOutput
from uuid import uuid4
from slugify import slugify
from sand.plugins.drepr.resources import get_entity_resource, get_table_resource
from sand.plugins.drepr.semanticmodel import get_drepr_sm, get_entity_data_nodes
from sand.plugins.drepr.transformation import has_transformation, get_transformation


def create_dataset_model(table: Table, sm: O.SemanticModel) -> DRepr:
    """Create a D-REPR model of the dataset."""
    columns = [slugify(c).replace("-", "_") for c in table.columns]
    get_attr_id = lambda ci: f"{ci}__{columns[ci]}"
    get_ent_attr_id = lambda ci: f"{ci}__ent__{columns[ci]}"
    ent_dnodes = get_entity_data_nodes(sm)

    attrs = [
        Attr(
            id=get_attr_id(ci),
            resource_id="table",
            path=Path(
                steps=[
                    RangeExpr(start=1, end=table.size + 1, step=1),
                    IndexExpr(val=ci),
                ]
            ),
            missing_values=[""],
        )
        for ci in range(len(table.columns))
    ]
    attrs += [
        Attr(
            id=get_ent_attr_id(node.col_index),
            resource_id="entity",
            path=Path(
                steps=[
                    RangeExpr(start=1, end=table.size + 1, step=1),
                    IndexExpr(val=node.col_index),
                ]
            ),
            missing_values=[""],
        )
        for node in ent_dnodes
    ]

    id2props = ChainedMapping(
        OntPropertyAR(),
        M.import_attr(SETTINGS["ont_props"]["default"]),
    )
    dsm = get_drepr_sm(sm, id2props, get_attr_id, get_ent_attr_id)

    datatype_transformations = []
    for node in sm.nodes():
        if not isinstance(node, O.DataNode):
            continue

        datatypes: Set[OntPropertyDataType] = {
            id2props[OntProperty.uri2id(inedge.abs_uri)].datatype
            for inedge in sm.in_edges(node.id)
        }
        datatype = list(datatypes)[0] if len(datatypes) == 1 else None
        if datatype is None or not has_transformation(datatype):
            continue
        datatype_transformations.append(
            Preprocessing(
                type=PreprocessingType.pmap,
                value=PMap(
                    resource_id="table",
                    path=Path(
                        steps=[
                            RangeExpr(start=1, end=table.size + 1, step=1),
                            IndexExpr(val=node.col_index),
                        ]
                    ),
                    code=get_transformation(datatype),
                    change_structure=False,
                ),
            )
        )

    return DRepr(
        resources=[
            Resource(id="table", type=ResourceType.CSV, prop=CSVProp()),
            Resource(id="entity", type=ResourceType.CSV, prop=CSVProp()),
        ],
        preprocessing=datatype_transformations,
        attrs=attrs,
        aligns=[
            RangeAlignment(
                source=get_attr_id(0),
                target=get_attr_id(ci),
                aligned_steps=[AlignedStep(source_idx=0, target_idx=0)],
            )
            for ci in range(1, len(table.columns))
        ]
        + [
            RangeAlignment(
                source=get_attr_id(0),
                target=get_ent_attr_id(node.col_index),
                aligned_steps=[AlignedStep(source_idx=0, target_idx=0)],
            )
            for node in ent_dnodes
        ],
        sm=dsm,
    )


def relational2rdf(table: Table, rows: List[TableRow], sm: O.SemanticModel):
    """Convert a relational table into RDF format"""
    if len(table.columns) == 0:
        # no column, no data
        return ""

    ent_columns = {node.col_index for node in get_entity_data_nodes(sm)}
    resources = {
        "table": get_table_resource(table, rows),
        "entity": get_entity_resource(table, rows, ent_columns),
    }

    content = execute(
        ds_model=create_dataset_model(table, sm),
        resources=resources,
        output=MemoryOutput(OutputFormat.TTL),
        debug=False,
    )
    return content
